# Część 8: Ładowanie modeli trójwymiarowych

## Wczytywanie modelu z pliku .obj

1. Utwórz nowy folder w swoim środowisku. Skopiuj do niego końcowe wersje plików z części 7, które będą potrzebne do kontynuowania projektu. Powinny się w nim znaleźć m.in.: `Button.py`, `Cube.py`, `Mesh3D.py`, `Object.py`, `Settings.py`, `Transform.py` oraz `Utils.py`. Dodatkowo skopiuj z części 7 plik głównego programu `AddingButtons.py` i zmień jego nazwę na **DisplayTeapot.py** (będzie to nasz program testowy w części 8).

2. W folderze utwórz podfolder o nazwie **models**. Pobierz plik modelu 3D **teapot.obj** (trójwymiarowy model czajnika) i umieść ten plik w nowo utworzonym folderze `models`. Ten plik posłuży nam do załadowania triangulowanego modelu w aplikacji.

3. Utwórz nowy skrypt Pythona o nazwie **LoadMesh.py**. Będzie on zawierał klasę ładującą model z pliku `.obj`. Wypełnij plik **LoadMesh.py** poniższym kodem źródłowym:

   ```python
   from Mesh3D import *
   
   class LoadMesh(Mesh3D):
       def __init__(self, draw_type, model_filename):
           self.vertices, self.triangles = self.load_drawing(model_filename)
           self.draw_type = draw_type
   
       def draw(self):
           for t in range(0, len(self.triangles), 3):
               glBegin(self.draw_type)
               glVertex3fv(self.vertices[self.triangles[t]])
               glVertex3fv(self.vertices[self.triangles[t+1]])
               glVertex3fv(self.vertices[self.triangles[t+2]])
               glEnd()
           glDisable(GL_TEXTURE_2D)
   
       def load_drawing(self, filename):
           vertices = []
           triangles = []
           with open(filename) as fp:
               line = fp.readline()
               while line:
                   if line[:2] == "v ":
                       vx, vy, vz = [float(value) for value in line[2:].split()]
                       vertices.append((vx, vy, vz))
                   if line[:2] == "f ":
                       t1, t2, t3 = [value for value in line[2:].split()]
                       triangles.append([int(value) for value in t1.split('/')][0] - 1)
                       triangles.append([int(value) for value in t2.split('/')][0] - 1)
                       triangles.append([int(value) for value in t3.split('/')][0] - 1)
                   line = fp.readline()
           return vertices, triangles
   ```

4. Plik **DisplayTeapot.py** zmodyfikuj go tak, aby wykorzystać nowo utworzoną klasę `LoadMesh`. W nagłówku pliku zaimportuj moduł `LoadMesh` (dodaj linię `from LoadMesh import *`). Następnie odszukaj miejsce, w którym tworzony jest obiekt `cube` i dodawany jest do niego komponent `Cube` – zamienimy ten komponent na `LoadMesh`. Kod w pliku **DisplayTeapot.py** powinien zostać zmieniony tak, aby wyglądał następująco:

   ```python
   from Object import *
   from Cube import *
   from LoadMesh import *
   ...
   objects_2d = []
   cube = Object("Cube")
   cube.add_component(Transform((0, 0, -5)))
   cube.add_component(LoadMesh(GL_LINE_LOOP, "models/teapot.obj"))
   objects_3d.append(cube)
   ```


## Wyjaśnienie komponentu LoadMesh

W pliku **LoadMesh.py** zdefiniowaliśmy klasę `LoadMesh`, która dziedziczy po istniejącej klasie `Mesh3D`. Jej zadaniem jest wczytanie modelu z zewnętrznego pliku `.obj` i narysowanie go w naszej aplikacji:

- Konstruktor `__init__` klasy `LoadMesh` przyjmuje parametry `draw_type` oraz `model_filename`. `draw_type` to stała OpenGL określająca tryb rysowania (np. `GL_LINE_LOOP` do rysowania krawędzi trójkątów lub `GL_TRIANGLES` do rysowania wypełnionych trójkątów). `model_filename` to ścieżka do pliku modelu (.obj). W konstruktorze wywołujemy metodę `self.load_drawing(model_filename)`, która zwraca listę wierzchołków i listę trójkątów odczytane z pliku – przypisujemy je do `self.vertices` i `self.triangles`. Następnie ustawiamy `self.draw_type` zgodnie z parametrem.

- Metoda `load_drawing(self, filename)` otwiera wskazany plik `.obj` i odczytuje z niego dane modeli. Format **Wavefront OBJ** reprezentuje wierzchołki oraz ściany (trójkąty) za pomocą linii tekstu:
  - Linia zaczynająca się od `"v "` oznacza współrzędne wierzchołka. Przykładowo linia `v 0.0 2.4 -1.4` oznacza wierzchołek o współrzędnych *(0.0, 2.4, -1.4)*. W kodzie sprawdzamy `if line[:2] == "v ":` i wtedy dokonujemy podziału tekstu, konwertujemy trzy wartości na typ float i dodajemy krotkę `(vx, vy, vz)` do listy `vertices`.
  - Linia zaczynająca się od `"f "` opisuje ścianę modelu przez indeksy wierzchołków (face). Przykładowo `f 1/1/1 2/2/2 3/3/3` oznacza trójkąt złożony z wierzchołków o indeksach 1, 2, 3 (dodatkowo w pliku po ukośnikach mogą znajdować się indeksy współrzędnych tekstury i normalnych, które na razie pomijamy). W naszym kodzie odczytujemy taką linię, dzielimy ją, i z każdej grupy typu `M/N/O` bierzemy pierwszą liczbę (reprezentującą indeks wierzchołka). Następnie odejmujemy 1 (ponieważ indeksy w pliku .obj zaczynają się od 1, a w Pythonie lista ma indeksy od 0) i zapisujemy do listy `triangles`. Zapisujemy kolejno trzy indeksy wierzchołków każdego trójkąta.

  W wyniku działania `load_drawing` otrzymujemy:
  - `vertices` – listę wszystkich wierzchołków modelu (każdy w postaci krotki trzech współrzędnych).
  - `triangles` – listę indeksów składającą się z grup po 3 indeksy. Każde kolejne trójki indeksów określają jeden trójkąt (np. `[a, b, c]` oznacza trójkąt z wierzchołków `vertices[a]`, `vertices[b]`, `vertices[c]`).

- Metoda `draw(self)` klasy `LoadMesh` jest nadpisaniem (override) metody z klasy bazowej `Mesh3D`. Iteruje ona po liście `self.triangles` co 3 indeksy (zmienna `t` przyjmuje wartości 0, 3, 6, ... aż do końca listy). Dla każdej takiej trójki indeksów wywołujemy sekwencję funkcji OpenGL, aby narysować trójkąt:
  - `glBegin(self.draw_type)` – rozpoczyna rysowanie prymitywu graficznego zadanego typu (w naszym przypadku `GL_LINE_LOOP`, co oznacza, że OpenGL połączy linie między podanymi wierzchołkami i zamknie pętlę, tworząc kontur trójkąta).
  - `glVertex3fv(self.vertices[index])` – dla każdego z trzech wierzchołków trójkąta przekazujemy współrzędne wierzchołka do potoku graficznego. Funkcja ta pobiera wektor 3-elementowy (x, y, z) i traktuje go jako kolejny wierzchołek rysowanej figury.
  - Po podaniu trzech wierzchołków wywołujemy `glEnd()`, co kończy rysowanie aktualnego prymitywu (tu: jednego trójkąta w formie linii).
  - Po narysowaniu wszystkich trójkątów wywołujemy `glDisable(GL_TEXTURE_2D)`. Ta linia dezaktywuje ewentualne mapowanie tekstur. W naszym przypadku model jest rysowany bez tekstury (jako siatka), więc wyłączenie teksturowania zapewnia, że żadne tekstury nie wpływają na renderowanie (linia ta została przeniesiona z klasy bazowej, gdzie mogła służyć do czyszczenia stanu po narysowaniu obiektu teksturowanego).